package be.iminds.ilabt.jfed.experimenter_gui.dialogs;

import be.iminds.ilabt.jfed.experimenter_gui.config.JFedGuiConfig;
import be.iminds.ilabt.jfed.experimenter_gui.ext_requests.JFedEnvironment;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Server;
import be.iminds.ilabt.jfed.fedmon.webapi.service.json.Testbed;
import be.iminds.ilabt.jfed.highlevel.model.SfaModel;
import be.iminds.ilabt.jfed.highlevel.model.Slice;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.ListSubAuthoritiesTask;
import be.iminds.ilabt.jfed.highlevel.tasks.ListSubAuthorityMembersTask;
import be.iminds.ilabt.jfed.lowlevel.GeniUserProvider;
import be.iminds.ilabt.jfed.lowlevel.api_wrapper.UserAndSliceApiWrapper;
import be.iminds.ilabt.jfed.lowlevel.authority.AuthorityFinder;
import be.iminds.ilabt.jfed.preferences.GuiPreferenceKey;
import be.iminds.ilabt.jfed.preferences.JFedGuiPreferences;
import be.iminds.ilabt.jfed.rspec.model.ModelRspecType;
import be.iminds.ilabt.jfed.rspec.model.javafx_impl.FXModelRspec;
import be.iminds.ilabt.jfed.ui.javafx.GlyphUtils;
import be.iminds.ilabt.jfed.ui.javafx.dialogs.JFDialogs;
import be.iminds.ilabt.jfed.util.GeniUrn;
import javafx.application.Platform;
import javafx.beans.InvalidationListener;
import javafx.beans.Observable;
import javafx.beans.binding.Bindings;
import javafx.beans.binding.StringBinding;
import javafx.beans.property.ListProperty;
import javafx.beans.property.ReadOnlyListProperty;
import javafx.beans.property.SimpleListProperty;
import javafx.beans.value.ChangeListener;
import javafx.beans.value.ObservableValue;
import javafx.collections.*;
import javafx.event.ActionEvent;
import javafx.event.EventHandler;
import javafx.fxml.FXML;
import javafx.geometry.Pos;
import javafx.scene.control.*;
import javafx.scene.input.MouseEvent;
import javafx.scene.layout.GridPane;
import javafx.scene.layout.HBox;
import javafx.scene.layout.RowConstraints;
import javafx.scene.layout.VBox;
import javafx.scene.paint.Color;
import javafx.util.StringConverter;
import org.controlsfx.control.textfield.TextFields;
import org.controlsfx.glyphfont.FontAwesome;
import org.controlsfx.validation.ValidationMessage;
import org.controlsfx.validation.ValidationResult;
import org.controlsfx.validation.ValidationSupport;
import org.controlsfx.validation.Validator;
import org.controlsfx.validation.decoration.GraphicValidationDecoration;
import org.reactfx.AwaitingEventStream;
import org.reactfx.EventStreams;
import org.reactfx.Subscription;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import javax.inject.Inject;
import java.time.Duration;
import java.time.Instant;
import java.time.ZonedDateTime;
import java.util.*;
import java.util.concurrent.ConcurrentHashMap;
import java.util.stream.Collectors;
import java.util.stream.IntStream;

/**
 * User: twalcari
 * Date: 2/24/14
 * Time: 4:03 PM
 */
public class StartExperimentController extends GridPane {
    private static final Logger LOG = LoggerFactory.getLogger(StartExperimentController.class);

    private static final Duration MINIMUM_SLICE_DURATION = Duration.ofMinutes(1);
    private static final int MAX_DURATION_DAYS = 100;
    private final HighLevelTaskFactory hltf;
    private final GeniUserProvider geniUserProvider;
    private final SfaModel sfaModel;
    private final JFedEnvironment environment;
    private final JFedGuiConfig config;
    private final JFedGuiPreferences preferences;
    @Nonnull private final AuthorityFinder authorityFinder;
    @Nonnull private final GDPRDialogFactory gdprDialogFactory;


    private final ObservableMap<String, Slice> slices = FXCollections.observableHashMap();
    private final ListChangeListener<Slice> sliceListChangeListener = c -> {
        while (c.next()) {
            c.getAddedSubList().forEach(slice -> slices.put(slice.getName(), slice));
            c.getRemoved().forEach(slice -> slices.remove(slice.getName()));
        }
    };

    private int maximumSliceNameLength = 8;
    private Collection<Testbed> gdprTestbeds = Collections.emptyList();

    private int[] durationOfNonExistingSlice = null;


    @FXML
    protected TextField nameTextField;
    @FXML
    protected Button showExistingSlicesButton;
    @FXML
    protected ComboBox<Integer> hoursComboBox;
    @FXML
    protected ComboBox<Integer> minutesComboBox;
    @FXML
    protected ComboBox<Integer> daysComboBox;
    @FXML
    protected Label subAuthLabel;
    @FXML
    protected CheckBox subAuthCheckBox;
    @FXML
    protected ProgressBar subAuthFetchProgressBar;
    @FXML
    protected ComboBox<String> subAuthComboBox;
    @FXML
    protected Button refreshSubAuthButton;
    @FXML
    protected CheckBox autoshareCheckBox;
    @FXML
    protected CheckBox requestSlaCheckBox;
    private int userSetDays = 0, userSetHours = 2, userSetMinutes = 0;

    @FXML VBox gdprList;

    private final InvalidationListener expirationDateTimeInvalidationListener = new InvalidationListener() {
        @Override
        public void invalidated(Observable observable) {
            userSetDays = daysComboBox.getValue();
            userSetHours = hoursComboBox.getValue();
            userSetMinutes = minutesComboBox.getValue();
        }
    };
    private final ValidationSupport validationSupport;
    private final boolean canRequestProjectMemberSshKeys;
    private boolean prevRequestProjectMemberSshKeysValue;

//    private final Map<String, ListProperty<GeniUrn>> projectNameToUsers = new ConcurrentHashMap<>();

    private boolean autoSetShareCheckbox = true;  //becomes false once user manually sets the checkbox

    @Inject
    StartExperimentController(HighLevelTaskFactory hltf, GeniUserProvider geniUserProvider,
                              SfaModel sfaModel, JFedEnvironment environment, JFedGuiConfig config, JFedGuiPreferences preferences,
                              @Nonnull AuthorityFinder authorityFinder, @Nonnull GDPRDialogFactory gdprDialogFactory) {
        this.hltf = hltf;
        this.geniUserProvider = geniUserProvider;
        this.sfaModel = sfaModel;
        this.environment = environment;
        this.config = config;
        this.preferences = preferences;
        this.authorityFinder = authorityFinder;
        this.gdprDialogFactory = gdprDialogFactory;

        validationSupport = new ValidationSupport();
        validationSupport.setValidationDecorator(new GraphicValidationDecoration());

        Server userServer =
                this.geniUserProvider.getLoggedInGeniUser().getUserAuthorityServer();
        assert userServer != null;
        canRequestProjectMemberSshKeys = userServer.hasFlag(Server.Flag.featureFedMaRequestMemberSshKey);
    }

    private Subscription sliceNameTextSubsciption;


    @FXML
    public void initialize() {
        //populate date/time combo-boxes

        //check if there is a startup parameter that defines the default duration, otherwise use the values set in userSet{days, hours, minutes}
        if (environment.getDefaultSliceDuration() != null) {
            int defaultSliceDuration = environment.getDefaultSliceDuration();

            userSetMinutes = defaultSliceDuration % 60;
            defaultSliceDuration /= 60;
            userSetHours = defaultSliceDuration % 24;
            userSetDays = Math.min(defaultSliceDuration / 24, MAX_DURATION_DAYS);
        }


        daysComboBox.getItems().addAll(IntStream.range(0, MAX_DURATION_DAYS).boxed().collect(Collectors.toList()));
        daysComboBox.getSelectionModel().select(userSetDays);

        hoursComboBox.getItems().addAll(IntStream.range(0, 24).boxed().collect(Collectors.toList()));
        hoursComboBox.getSelectionModel().select(userSetHours);

        minutesComboBox.getItems().addAll(IntStream.range(0, 60).boxed().collect(Collectors.toList()));
        minutesComboBox.setConverter(new StringConverter<Integer>() {
            @Override
            public String toString(Integer integer) {
                return String.format("%02d", integer);
            }

            @Override
            public Integer fromString(String s) {
                return Integer.parseInt(s);
            }
        });
        minutesComboBox.getSelectionModel().select(userSetMinutes);

        //make our slices-variable reflect the currently known slices
        sfaModel.getSlices().forEach(slice -> slices.put(slice.getName(), slice));
        sfaModel.getSlices().addListener(new WeakListChangeListener<>(sliceListChangeListener));

        TextFields.bindAutoCompletion(nameTextField, slices.keySet());

        subAuthCheckBox.selectedProperty().addListener(observable -> {
            if (subAuthCheckBox.isSelected())
                enableSubAuthoritiesList();
            else
                disableSubAuthoritiesList();
        });
        subAuthCheckBox.selectedProperty().addListener(observable -> {
            if (!subAuthCheckBox.isSelected()) {
                prevRequestProjectMemberSshKeysValue = autoshareCheckBox.isSelected();
                autoshareCheckBox.setSelected(false);

                autoshareCheckBox.setDisable(true);
            } else {
                autoshareCheckBox.setSelected(prevRequestProjectMemberSshKeysValue);
                autoshareCheckBox.setDisable(false);
            }
        });

        subAuthComboBox.getSelectionModel().selectedItemProperty().addListener((observable, oldProjectName, newProjectName) -> {
            if (newProjectName != null) {
                if (autoSetShareCheckbox)
                    setAutoShareForProject(newProjectName);
                ReadOnlyListProperty<GeniUrn> userUrns = sfaModel.getUsersBySubAuthority(newProjectName);
//                subAuthUserCountLabel.visibleProperty().bind(userUrns.emptyProperty().not());
                StringBinding userCountTextBinding = new StringBinding() {
                    {
                        super.bind(userUrns.sizeProperty());
                    }
                    @Override
                    protected String computeValue() {
                        if (userUrns.isEmpty())
                            return "Share Experiment with all Project Members";
                        else
                            return "Share Experiment with all "+userUrns.sizeProperty().get()+" Project Members";
                    }
                }; //Share Experiment with all Project Members
                autoshareCheckBox.textProperty().bind(userCountTextBinding);
            } else {
                autoshareCheckBox.setSelected(false);
//                subAuthUserCountLabel.visibleProperty().unbind();
                autoshareCheckBox.textProperty().unbind();
                autoshareCheckBox.setText("Share Experiment with all Project Members");
            }
        });

        enableSliceDetailsEditing();



        AwaitingEventStream<Void> sliceNameEventStream = EventStreams.invalidationsOf(nameTextField.textProperty()).reduceSuccessions((a, b) -> b, Duration.ofMillis(350));
        sliceNameTextSubsciption = sliceNameEventStream.subscribe(o -> {
            String sliceName = nameTextField.getText();

            if (slices.containsKey(sliceName)) {
                if(durationOfNonExistingSlice == null){
                    durationOfNonExistingSlice = new int[]{userSetDays, userSetHours, userSetMinutes};
                }

                showFixedSliceDetails(slices.get(sliceName));
            } else {
                enableSliceDetailsEditing();

                if(durationOfNonExistingSlice != null){
                    daysComboBox.setValue(durationOfNonExistingSlice[0]);
                    hoursComboBox.setValue(durationOfNonExistingSlice[1]);
                    minutesComboBox.setValue(durationOfNonExistingSlice[2]);

                    durationOfNonExistingSlice = null;
                }
            }
        });

        Platform.runLater(nameTextField::requestFocus);


        autoshareCheckBox.managedProperty().bind(autoshareCheckBox.visibleProperty());

        autoshareCheckBox.setVisible(canRequestProjectMemberSshKeys);
        autoshareCheckBox.setSelected(false);
//     was:   autoshareCheckBox.setSelected(canRequestProjectMemberSshKeys && preferences.getAddProjectMembersSshKeysByDefault());
        autoshareCheckBox.setOnAction((e) -> autoSetShareCheckbox = false);

        requestSlaCheckBox.setVisible(config.isSlaAvailable());

        validationSupport.registerValidator(nameTextField, false, sliceNameValidator);
        validationSupport.<String>registerValidator(subAuthComboBox,
                Validator.createPredicateValidator(
                        s -> !subAuthCheckBox.isSelected() || !subAuthComboBox.getSelectionModel().isEmpty(),
                        "Please choose a project in which this experiment should be run!"));


        addGdprLinesWhenNeeded(gdprTestbeds);
    }


    @Nonnull
    private GeniUrn getProjectUrn(@Nonnull String projectName) {
        return GeniUrn.createGeniUrnFromEncodedParts(
                geniUserProvider.getLoggedInGeniUser().getUserUrn().getEncodedTopLevelAuthority_withoutSubAuth(),
                "project",
                projectName);
    }

    private void setAutoShareForProject(@Nonnull String projectName) {
        Boolean autoshare = preferences.getExperimentAutoshare(getProjectUrn(projectName));
        if (autoshare == null) {
            int userCount = sfaModel.getUsersBySubAuthority(projectName).size();
            if (userCount > 0 && userCount <= 5)
                autoshareCheckBox.setSelected(true);
            else
                autoshareCheckBox.setSelected(false);
        } else {
            autoshareCheckBox.setSelected(canRequestProjectMemberSshKeys && autoshare);
        }
    }
    public void saveAutoShare() {
        String projectName = subAuthComboBox.getSelectionModel().getSelectedItem();
        if (projectName != null) {
            boolean autoshare = autoshareCheckBox.isSelected();
            preferences.setExperimentAutoshare(getProjectUrn(projectName), autoshare);
        }
    }
    public void saveLastProject() {
        String projectName = subAuthComboBox.getSelectionModel().getSelectedItem();
        preferences.setString(GuiPreferenceKey.PREF_EXPERIMENT_LAST_PROJECT, projectName);
    }

    private boolean sliceExpirationEditingEnabled = false;

    //is set to false, so that enableSliceDetailsEditing() would successfully run the first time it is called
    private boolean sliceDetailsEditingEnabled = false;

    private void enableSliceDetailsEditing() {
        if (sliceDetailsEditingEnabled)
            return;

        initializeSubAuthorities();

        sliceDetailsEditingEnabled = true;
    }

    private void disableSliceDetailsEditing() {
        if (!sliceDetailsEditingEnabled)
            return;

        subAuthCheckBox.setDisable(true);
        subAuthComboBox.setDisable(true);

        sliceDetailsEditingEnabled = false;
    }

    private void enableSliceExpirationEditing() {
        if (sliceExpirationEditingEnabled)
            return;

        daysComboBox.setDisable(false);
        hoursComboBox.setDisable(false);
        minutesComboBox.setDisable(false);

        daysComboBox.setValue(userSetDays);
        hoursComboBox.setValue(userSetHours);
        minutesComboBox.setValue(userSetMinutes);

        daysComboBox.valueProperty().addListener(expirationDateTimeInvalidationListener);
        hoursComboBox.valueProperty().addListener(expirationDateTimeInvalidationListener);
        minutesComboBox.valueProperty().addListener(expirationDateTimeInvalidationListener);

        sliceExpirationEditingEnabled = true;
    }

    private void disableSliceExpirationEditing() {
        if (!sliceExpirationEditingEnabled)
            return;

        daysComboBox.valueProperty().removeListener(expirationDateTimeInvalidationListener);
        hoursComboBox.valueProperty().removeListener(expirationDateTimeInvalidationListener);
        minutesComboBox.valueProperty().removeListener(expirationDateTimeInvalidationListener);

        daysComboBox.setDisable(true);
        hoursComboBox.setDisable(true);
        minutesComboBox.setDisable(true);

        sliceExpirationEditingEnabled = false;
    }

    private void showFixedSliceDetails(Slice slice) {
        if (sliceDetailsEditingEnabled) {
            disableSliceDetailsEditing();
        }

        enableSliceExpirationEditing();

        if (slice.getSubAuthorityName() != null) {
            subAuthCheckBox.setSelected(true);
            subAuthComboBox.setValue(slice.getSubAuthorityName());
        } else {
            subAuthCheckBox.setSelected(false);
            subAuthComboBox.setValue(null);
        }

        //fix the expiration-time to that of the slice
        if (slice.getExpirationDate() != null) {
            Duration duration = Duration.between(Instant.now(), slice.getExpirationDate());

            if(Duration.ofHours(2).compareTo(duration) < 0) {

                long time = duration.toMinutes();

                minutesComboBox.setValue((int) (time % 60));
                time /= 60;

                hoursComboBox.setValue((int) (time % 24));
                time /= 24;

                daysComboBox.setValue((int) time);
            }
        }// else {
        //do nothing, value will already be initialised

//            daysComboBox.getSelectionModel().clearSelection();
//            hoursComboBox.getSelectionModel().clearSelection();
//            minutesComboBox.getSelectionModel().clearSelection();
        //  }
    }

    private final Validator<String> sliceNameValidator = (control, sliceName) -> {
        ValidationResult result = new ValidationResult();
        if (sliceName == null) {
            result.add(ValidationMessage.error(control, "Experiment name cannot be empty!"));
            return result;
        }

        if (slices.containsKey(sliceName)) {
            //always allow existing slices!
            return result;
        }

        if (sliceName.length() == 0) {
            result.add(ValidationMessage.error(control, "Experiment name cannot be empty!"));
        } else {
            if (sliceName.length() > maximumSliceNameLength) {
                if (maximumSliceNameLength > 5)
                    result.add(ValidationMessage.error(control,
                            String.format("Experiment name cannot be longer than %d characters", maximumSliceNameLength)));
                else {
                    if (maximumSliceNameLength <= 0)
                        result.add(ValidationMessage.error(control,
                                String.format("Your nodes names are too long!" +
                                        "Try to remove at least %d chars from the longest node name(s)", (3 - maximumSliceNameLength))));
                    else
                        result.add(ValidationMessage.error(control,
                                String.format("Experiment name cannot be longer than %d characters. You can try to shorten node names to increase this limit.", maximumSliceNameLength)));
                }
            }
        }

        for (char c : sliceName.toCharArray()) {
            boolean validChar = (c >= 'a' && c <= 'z')
                    || (c >= 'A' && c <= 'Z')
                    || (c >= '0' && c <= '9');

            if (!validChar) {
                result.add(ValidationMessage.error(control,
                        String.format("Experiment name may not contain the character '%c'." +
                                " Only characters [a-zA-Z0-9] are allowed", c)));
            }
        }

        return result;
    };

    private void initializeSubAuthorities() {
        //initialize subauthorities-list
        if (sfaModel.getSubAuthoritySupport()
                == UserAndSliceApiWrapper.SubAuthoritySupport.SUB_AUTHORITY_FORBIDDEN) {
            subAuthCheckBox.setSelected(false);
            subAuthCheckBox.setDisable(true);

            subAuthFetchProgressBar.setVisible(false);
            subAuthComboBox.setDisable(true);
            subAuthComboBox.setItems(FXCollections.observableArrayList("Not supported"));
            refreshSubAuthButton.setDisable(true);
        } else {

            subAuthCheckBox.setSelected(true);
            if (sfaModel.getSubAuthoritySupport()
                    == UserAndSliceApiWrapper.SubAuthoritySupport.SUB_AUTHORITY_MANDATORY) {
                subAuthCheckBox.setDisable(true);
            } else {
                subAuthCheckBox.setDisable(false);
                //default: use a sub authority (because it is generally nicer)
            }
            enableSubAuthoritiesList();
        }
    }

//    private void clearAllProjectUsers() {
//        for (ListProperty<GeniUrn> observableList: projectNameToUsers.values()) {
//            observableList.clear();
//        }
//    }

    private void refreshProjectUsers(Collection<String> projectNames) {
//        clearAllProjectUsers();

        for (String projectName: projectNames) {
            ListSubAuthorityMembersTask listSubAuthorityMembersTask = this.hltf.listSubAuthorityMembers(projectName);
//            listSubAuthorityMembersTask.addCallback((task, taskExecution, state) -> {
//                List<GeniUrn> members = listSubAuthorityMembersTask.getMemberUrns();
//                Platform.runLater(() -> {
//                    getProjectUsers(projectName).setAll(members);
//                });
//            });
            this.hltf.submitTask(listSubAuthorityMembersTask);
        }
    }

//    /**
//     * @param projectName
//     * @return a list with the urns of the users of the project. If the list is empty, no data is known (yet)
//     */
//    @Nonnull
//    private ListProperty<GeniUrn> getProjectUsers(@Nonnull String projectName) {
//        return projectNameToUsers.computeIfAbsent(projectName, (k) -> new SimpleListProperty<>(FXCollections.observableArrayList()));
//    }

    private void showSubAuthoritiesList() {
        //this method should not be executed when SubAuthority-checkbox is not selected
        if (!subAuthCheckBox.isSelected()) {
            LOG.warn("unexpected: showSubAuthoritiesList() called while subAuthCheckBox.isSelected()=" + !subAuthCheckBox.isSelected());
            return;
        }

        List<String> subAuthNames = new ArrayList<>(sfaModel.getLoggedInUserSubAuthorityNames());
        Collections.sort(subAuthNames, String.CASE_INSENSITIVE_ORDER);

        subAuthComboBox.setItems(FXCollections.observableList(subAuthNames));
//        Collections.sort(subAuthComboBox.getItems());

        subAuthComboBox.setDisable(false);

        if (!subAuthComboBox.getItems().isEmpty()) {
            String selectedProject = null;

            String defaultProject = preferences.getString(GuiPreferenceKey.PREF_EXPERIMENT_DEFAULT_PROJECT);
            if (defaultProject != null) {
                selectedProject = defaultProject;
            } else {
                String lastProject = preferences.getString(GuiPreferenceKey.PREF_EXPERIMENT_LAST_PROJECT);
                selectedProject = lastProject;
            }

            boolean containsSelectedProject;
            if (selectedProject == null) {
                containsSelectedProject = false;
            } else {
                containsSelectedProject = false;
                for (String projectName: subAuthComboBox.getItems()) {
                    if (projectName.trim().equalsIgnoreCase(selectedProject.trim())) {
                        containsSelectedProject = true;
                        break;
                    }
                }
            }
            if (!containsSelectedProject) {
                subAuthComboBox.getSelectionModel().selectFirst();
            } else {
                subAuthComboBox.getSelectionModel().select(selectedProject);
            }
        }

        if (subAuthComboBox.getItems().isEmpty() && sfaModel.getSubAuthoritySupport() == UserAndSliceApiWrapper.SubAuthoritySupport.SUB_AUTHORITY_MANDATORY)
            JFDialogs.create()
                    .owner(subAuthComboBox)
                    .message("Selecting a project is mandatory for your experiment, but jFed did not receive any candidates." +
                            " Please contact your support team.")
                    .masthead("No projects available")
                    .showError();

        subAuthFetchProgressBar.setVisible(false);

        refreshSubAuthButton.setDisable(false);
    }

    private void disableSubAuthoritiesList() {
        subAuthLabel.setDisable(true);
        subAuthComboBox.getSelectionModel().clearSelection();
        subAuthComboBox.getItems().clear();
        subAuthComboBox.setDisable(true);

        subAuthFetchProgressBar.setVisible(false);

        refreshSubAuthButton.setDisable(true);
    }

    private void enableSubAuthoritiesList() {
        subAuthLabel.setDisable(false);

        if (shouldRefreshSubAuthorities()) {
            refreshSubAuthoritiesList();
        } else {
            showSubAuthoritiesList();
        }
    }

    private boolean shouldRefreshSubAuthorities() {
        ListSubAuthoritiesTask listSubAuthoritiesTask =
                hltf.listSubAuthorities();
        return listSubAuthoritiesTask.newExecutionNeeded();
    }

    @FXML
    private void refreshSubAuthoritiesList() {
        //already do this here, to show that this data too will be refreshed
        sfaModel.clearUsersBySubAuthority();

        subAuthComboBox.setItems(FXCollections.emptyObservableList());
        subAuthComboBox.setDisable(true);

        subAuthFetchProgressBar.setVisible(true);

        refreshSubAuthButton.setDisable(true);

        //initialize task to fetch sub-authority names
        ListSubAuthoritiesTask listSubAuthoritiesTask = hltf.listSubAuthorities();

        listSubAuthoritiesTask.addCallback((task, taskExecution, state) -> {
            Platform.runLater(this::showSubAuthoritiesList);
            refreshProjectUsers(sfaModel.getLoggedInUserSubAuthorityNames());
        });

        hltf.submitTask(listSubAuthoritiesTask);
    }

    @FXML
    private void showExistingSlicesButton() {
        Set<String> activeSliceNames = slices.keySet();

        if (!activeSliceNames.isEmpty()) {
            ChoiceDialog<String> existingSlicesDialog = new ChoiceDialog<>();
            existingSlicesDialog.setTitle("Select an existing slice");
            existingSlicesDialog.setHeaderText("Select an existing slice");
            existingSlicesDialog.getItems().addAll(activeSliceNames);

            existingSlicesDialog.showAndWait().ifPresent(nameTextField::setText);
        } else {

            JFDialogs.create()
                    .masthead("No slices found")
                    .title("No slices found")
                    .message("You currently don't have any active slices to select from. Please create a new slice to continue.")
                    .showInformation();
        }

    }



    public void addGdprLinesWhenNeeded(Collection<Testbed> testbeds) {
        if (testbeds.isEmpty())
            return;

        if (gdprList == null)
            return;

        gdprList.visibleProperty().setValue(!testbeds.isEmpty());
        gdprList.getChildren().clear();

        for (Testbed testbed : testbeds) {
            assert gdprList.isVisible();
            addTestbedGdprLine(testbed);
        }
        gdprList.requestLayout();
    }

    public void addTestbedGdprLine(@Nonnull final Testbed testbed) {
        LOG.debug("addTestbedGdprLine for "+testbed.getLongName());
        HBox child = new HBox();
        child.setAlignment(Pos.CENTER_LEFT);
        child.setSpacing(10.0);
        gdprList.getChildren().add(child);

        child.getChildren().add(GlyphUtils.createGlyph(FontAwesome.Glyph.EXCLAMATION_TRIANGLE, Color.ORANGERED));
        child.getChildren().add(new Label(testbed.getLongName()));
        child.getChildren().add(new Label("-"));
        Label consentLabel = new Label("Manage Consent");
        consentLabel.getStyleClass().add("label-link");
        consentLabel.setOnMouseClicked(new EventHandler<MouseEvent>() {
            @Override
            public void handle(MouseEvent event) {
                gdprDialogFactory.showDialog(testbed);
            }
        });
        child.getChildren().add(consentLabel);
        child.getChildren().add(new Label("-"));
        Label actionLabel = new Label("Action Required");
        actionLabel.getStyleClass().add("action-required");
        child.getChildren().add(actionLabel);
    }

    public ValidationSupport getValidationSupport() {
        return validationSupport;
    }

    public void setMaximumSliceNameLength(int maximumSliceNameLength) {
        this.maximumSliceNameLength = maximumSliceNameLength;
    }

    public void setGdprTestbeds(Collection<Testbed> gdprTestbeds) {
        this.gdprTestbeds = gdprTestbeds;
        addGdprLinesWhenNeeded(gdprTestbeds);
    }

    public String getSliceName() {
        return nameTextField.getText();
    }

    public String getSubAuthName() {
        return subAuthCheckBox.isSelected() ? subAuthComboBox.getValue() : null;
    }

    public Instant getExpirationTime() {
        if (daysComboBox.getSelectionModel().isEmpty() ||
                hoursComboBox.getSelectionModel().isEmpty() ||
                minutesComboBox.getSelectionModel().isEmpty())
            return null;


        return ZonedDateTime.now()
                .plusDays(daysComboBox.getSelectionModel().getSelectedItem())
                .plusHours(hoursComboBox.getSelectionModel().getSelectedItem())
                .plusMinutes(minutesComboBox.getSelectionModel().getSelectedItem())
                .plusSeconds(15).toInstant();
    }

    public boolean isAutoShareExperiment() {
        return autoshareCheckBox.isSelected();
    }

    public boolean isRequestSla() {
        return requestSlaCheckBox.isSelected();
    }

    public TextField getNameTextField() {
        return nameTextField;
    }
}
